/*    
 *    Important Note: The neutral conductor must be connected to the consumer star point!
 *                    Consumers in a delta connection must not be operated with this controller!
 *                    
 *                    In the following contend, phase 1 refers to the phase that is connected to the controller board power supply at terminal L.
 *                    Phase 2 and 3 are each offset by 120 degrees (clockwise rotating field).
 *                    Please keep this in mind, when connecting consumers to the burst firing outputs O1 to O3 and measure the rotating field before.
 *                    
 *                    You can either connect one 3-phase consumer, or a 2-phase and a 1-phase consumer, or three 1-phase consumers to this board.
 *                    In order for this to work, the phase(s) of each consumer to which it is connected must be assigned to the controller outputs.
 */

// Mains
#define MAINS_FREQUENCY 50             // [Hz] - Mains rated frequency (10 Hz ... 100 Hz)
#define MAXIMUM_FREQUENCY_DEVIATION 5  // [%]  - Maximum frequency deviaton from rated frequency (±2% ... ±10%) \
                                       //        A larger frequency deviation will stop the burst control

// Digital output O1
#define O1_CONNECTED_POWER 100       // [W]  - Rated Power of the connected consumer at output O1 (0 W ... 10000 W)
#define O1_ASSIGNED_TO_PHASE 1       // []   - Output O1 is assigned to phase (1 ... 3)
#define O1_OPERATING_MODE 0          // []   - Specifies the control type of the consumer load (0 = half sine waves / 1 = full sine waves)
#define O1_LOAD_PATTERN 0            // []   - Specifies the way in which the load is distributed over a load cycle (0 = asymmetrical / 1 = symmetrical)
#define O1_LOAD_STEP_RESOLUTION 100  // [%]  - Determines the relationship between load step cycle time and load step resolution (10% ... 1000%) \
                                     //        10% = Coarsest load step resolution, but fastest load step cycle time of 100 ms \
                                     //        100% = Standard load step resolution at a load step cycle time of 1 second \
                                     //        1000% = Finest possible load step resolution, but slowest load step cycle time of 10 seconds
// Digital output O2
#define O2_CONNECTED_POWER 100       // [W]  - Rated Power of the connected consumer at output O2 (0 W ... 10000 W)
#define O2_ASSIGNED_TO_PHASE 2       // []   - Output O2 is assigned to phase (1 ... 3)
#define O2_OPERATING_MODE 0          // []   - Specifies the control type of the consumer load (0 = half sine waves / 1 = full sine waves)
#define O2_LOAD_PATTERN 0            // []   - Specifies the way in which the load is distributed over a load cycle (0 = asymmetrical / 1 = symmetrical)
#define O2_LOAD_STEP_RESOLUTION 100  // [%]  - Determines the relationship between load step cycle time and load step resolution (10% ... 1000%) \
                                     //        10% = Coarsest load step resolution, but fastest load step cycle time of 100 ms \
                                     //        100% = Standard load step resolution at a load step cycle time of 1 second \
                                     //        1000% = Finest possible load step resolution, but slowest load step cycle time of 10 seconds
// Digital output O3
#define O3_CONNECTED_POWER 100      // [W]  - Rated Power of the connected consumer at output O3 (0 W ... 10000 W)
#define O3_ASSIGNED_TO_PHASE 3      // []   - Output O3 is assigned to phase (1 ... 3)
#define O3_OPERATING_MODE 0         // []   - Specifies the control type of the consumer load (0 = half sine waves / 1 = full sine waves)
#define O3_LOAD_PATTERN 0           // []   - Specifies the way in which the load is distributed over a load cycle (0 = asymmetrical / 1 = symmetrical)
#define O3_LOAD_STEP_RESOLUTION 100 // [%]  - Determines the relationship between load step cycle time and load step resolution (10% ... 1000%) \
                                    //        10% = Coarsest load step resolution, but fastest load step cycle time of 100 ms \
                                    //        100% = Standard load step resolution at a load step cycle time of 1 second \
                                    //        1000% = Finest possible load step resolution, but slowest load step cycle time of 10 seconds
// Relay output K1
#define K1_CONNECTED_POWER 100  // [W]  - Rated Power of the connected consumer at relay output K1 (0 W ... 10000 W)

// Relay output K2
#define K2_CONNECTED_POWER 100  // [W]  - Rated Power of the connected consumer at relay output K2 (0 W ... 10000 W)


portMUX_TYPE timerMux = portMUX_INITIALIZER_UNLOCKED;


/*  
This routines uses Hardware timer 0 and 1
*/

// I/O definition
#define ZERO_CROSSING_DETECTION_PHW 34  // Zero crossing detection input (positive half wave)
#define ZERO_CROSSING_DETECTION_NHW 35  // Zero crossing detection input (negative half wave)
#define O1_OUTPUT 25                    // Burst firing control output O1
#define O2_OUTPUT 26                    // Burst firing control output O2
#define O3_OUTPUT 27                    // Burst firing control output O3
#define K1_OUTPUT 18                    // Relay output K1
#define K2_OUTPUT 19                    // Relay output K2

// Fixed values
#define MINIMUM_LOAD_STEP_RESOLUTION 10    // [%] - Minimum allowed load step resolution
#define MAXIMUM_LOAD_STEP_RESOLUTION 1000  // [%] - Maximum allowed load step resolution (1000 = 10 seconds)
#define MAXIMUM_CYCLE_PATTERN_VARS 1000    // [n] - Maximum allowed cycle pattern variables (uint64_t) for each symmetrical burst control output
#define MINIMUM_CONSUMER_POWER 0           // [W] - Minimum allowed consumer power per output
#define MAXIMUM_CONSUMER_POWER 10000       // [W] - Maximum allowed consumer power per output
#define PHASE_TRIGGER_ANGLE -30            // [°] - Negative triac firing release angle before phase reaches zero crossing point
#define TIME_OUT_ANGLE 220                 // [°] - Half wave time out angle after which a main frequency fault is detected
#define MAINS_STABILIZATION_TIME 10        // [s] - Stabilization time after a mains frequency fault
#define REMAINING_WAVES 1                  // [n] - Number of (half) waves remaining when a new load value is requested
#define TIME_BASE 10000000ULL

// Configuration check
#if (MAINS_FREQUENCY < 10)
#error Mains rated frequency is to low.
#endif
#if (MAINS_FREQUENCY > 100)
#error Mains rated frequency is to high.
#endif
#if (MAXIMUM_FREQUENCY_DEVIATION < 2)
#error Maximum frequency deviation is set to low.
#endif
#if (MAXIMUM_FREQUENCY_DEVIATION > 10)
#error Maximum frequency deviation is set to high.
#endif
#if (O1_CONNECTED_POWER < MINIMUM_CONSUMER_POWER)
#error Consumer power at output O1 is to low.
#endif
#if (O1_CONNECTED_POWER > MAXIMUM_CONSUMER_POWER)
#error Consumer power at output O1 is to high.
#endif
#if (O2_CONNECTED_POWER < MINIMUM_CONSUMER_POWER)
#error Consumer power at output O2 is to low.
#endif
#if (O2_CONNECTED_POWER > MAXIMUM_CONSUMER_POWER)
#error Consumer power at output O2 is to high.
#endif
#if (O3_CONNECTED_POWER < MINIMUM_CONSUMER_POWER)
#error Consumer power at output O3 is to low.
#endif
#if (O3_CONNECTED_POWER > MAXIMUM_CONSUMER_POWER)
#error Consumer power at output O3 is to high.
#endif
#if (K1_CONNECTED_POWER < MINIMUM_CONSUMER_POWER)
#error Consumer power at relay output K1 is to low.
#endif
#if (K1_CONNECTED_POWER > MAXIMUM_CONSUMER_POWER)
#error Consumer power at relay output K1 is to high.
#endif
#if (K2_CONNECTED_POWER < MINIMUM_CONSUMER_POWER)
#error Consumer power at relay output K2 is to low.
#endif
#if (K2_CONNECTED_POWER > MAXIMUM_CONSUMER_POWER)
#error Consumer power at relay output K2 is to high.
#endif
#if ((O1_ASSIGNED_TO_PHASE != 1) && (O1_ASSIGNED_TO_PHASE != 2) && (O1_ASSIGNED_TO_PHASE != 3))
#error Illegal phase number assigned to output O1.
#endif
#if ((O2_ASSIGNED_TO_PHASE != 1) && (O2_ASSIGNED_TO_PHASE != 2) && (O2_ASSIGNED_TO_PHASE != 3))
#error Illegal phase number assigned to output O2.
#endif
#if ((O3_ASSIGNED_TO_PHASE != 1) && (O3_ASSIGNED_TO_PHASE != 2) && (O3_ASSIGNED_TO_PHASE != 3))
#error Illegal phase number assigned to output O3.
#endif
#if ((O1_OPERATING_MODE != 0) && (O1_OPERATING_MODE != 1))
#error Illegal operating mode is set to outout O1
#endif
#if ((O2_OPERATING_MODE != 0) && (O2_OPERATING_MODE != 1))
#error Illegal operating mode is set to outout O2
#endif
#if ((O3_OPERATING_MODE != 0) && (O3_OPERATING_MODE != 1))
#error Illegal operating mode is set to outout O3
#endif
#if ((O1_LOAD_PATTERN != 0) && (O1_LOAD_PATTERN != 1))
#error Illegal load pattern is set to outout O1
#endif
#if ((O2_LOAD_PATTERN != 0) && (O2_LOAD_PATTERN != 1))
#error Illegal load pattern is set to outout O2
#endif
#if ((O3_LOAD_PATTERN != 0) && (O3_LOAD_PATTERN != 1))
#error Illegal load pattern is set to outout O3
#endif
#if (O1_LOAD_STEP_RESOLUTION < MINIMUM_LOAD_STEP_RESOLUTION)
#error Load step resolution of output O1 is set to low.
#endif
#if (O1_LOAD_STEP_RESOLUTION > MAXIMUM_LOAD_STEP_RESOLUTION)
#error Load step resolution of output O1 is set to high.
#endif
#if (O2_LOAD_STEP_RESOLUTION < MINIMUM_LOAD_STEP_RESOLUTION)
#error Load step resolution of output O2 is set to low.
#endif
#if (O2_LOAD_STEP_RESOLUTION > MAXIMUM_LOAD_STEP_RESOLUTION)
#error Load step resolution of output O2 is set to high.
#endif
#if (O3_LOAD_STEP_RESOLUTION < MINIMUM_LOAD_STEP_RESOLUTION)
#error Load step resolution of output O3 is set to low.
#endif
#if (O3_LOAD_STEP_RESOLUTION > MAXIMUM_LOAD_STEP_RESOLUTION)
#error Load step resolution of output O3 is set to high.
#endif


// Calculated values
#define HALF_WAVES_PER_SECOND (MAINS_FREQUENCY << 1)
#define HALF_WAVE_TIME (TIME_BASE / HALF_WAVES_PER_SECOND)
#define TRIGGER_ANGLE_TIME_OUT (uint64_t)((HALF_WAVE_TIME / 180.0) * TIME_OUT_ANGLE)                // 220°
#define TRIGGER_ANGLE_PHASE_1 (uint64_t)((HALF_WAVE_TIME / 180.0) * (180.0 + PHASE_TRIGGER_ANGLE))  // 150°
#define TRIGGER_ANGLE_PHASE_2 (uint64_t)((HALF_WAVE_TIME / 180.0) * (120.0 + PHASE_TRIGGER_ANGLE))  // 90°
#define TRIGGER_ANGLE_PHASE_3 (uint64_t)((HALF_WAVE_TIME / 180.0) * (60.0 + PHASE_TRIGGER_ANGLE))   // 30°
#define HALF_WAVE_DURATION_MIN (uint64_t)(HALF_WAVE_TIME - (HALF_WAVE_TIME / 100.0 * MAXIMUM_FREQUENCY_DEVIATION))
#define HALF_WAVE_DURATION_MAX (uint64_t)(HALF_WAVE_TIME + (HALF_WAVE_TIME / 100.0 * MAXIMUM_FREQUENCY_DEVIATION))
#define MIN_FREQUENCY_DEVIATION (float)(MAINS_FREQUENCY - ((MAINS_FREQUENCY * MAXIMUM_FREQUENCY_DEVIATION) / 100.0))
#define MAX_FREQUENCY_DEVIATION (float)(MAINS_FREQUENCY + ((MAINS_FREQUENCY * MAXIMUM_FREQUENCY_DEVIATION) / 100.0))
#define MAINS_STABILIZATION_VALUE (HALF_WAVES_PER_SECOND * MAINS_STABILIZATION_TIME)
#if (O1_OPERATING_MODE == 0)
#define O1_WAVES_PER_CYCLE ((HALF_WAVES_PER_SECOND * O1_LOAD_STEP_RESOLUTION) / 100)   // (Half) waves per load step cycle at output 1
#define O1_LOAD_STEP_CYCLE_TIME ((1000 * O1_WAVES_PER_CYCLE) / HALF_WAVES_PER_SECOND)  // Load step cycle time at output 1
#else
#define O1_WAVES_PER_CYCLE ((MAINS_FREQUENCY * O1_LOAD_STEP_RESOLUTION) / 100)
#define O1_LOAD_STEP_CYCLE_TIME ((1000 * O1_WAVES_PER_CYCLE) / MAINS_FREQUENCY)
#endif
#if (O2_OPERATING_MODE == 0)
#define O2_WAVES_PER_CYCLE ((HALF_WAVES_PER_SECOND * O2_LOAD_STEP_RESOLUTION) / 100)
#define O2_LOAD_STEP_CYCLE_TIME ((1000 * O2_WAVES_PER_CYCLE) / HALF_WAVES_PER_SECOND)
#else
#define O2_WAVES_PER_CYCLE ((MAINS_FREQUENCY * O2_LOAD_STEP_RESOLUTION) / 100)
#define O2_LOAD_STEP_CYCLE_TIME ((1000 * O2_WAVES_PER_CYCLE) / MAINS_FREQUENCY)
#endif
#if (O3_OPERATING_MODE == 0)
#define O3_WAVES_PER_CYCLE ((HALF_WAVES_PER_SECOND * O3_LOAD_STEP_RESOLUTION) / 100)
#define O3_LOAD_STEP_CYCLE_TIME ((1000 * O3_WAVES_PER_CYCLE) / HALF_WAVES_PER_SECOND)
#else
#define O3_WAVES_PER_CYCLE ((MAINS_FREQUENCY * O3_LOAD_STEP_RESOLUTION) / 100)
#define O3_LOAD_STEP_CYCLE_TIME ((1000 * O3_WAVES_PER_CYCLE) / MAINS_FREQUENCY)
#endif
#define O1_LOAD_STEP_RESOLUTION_W ((float)O1_CONNECTED_POWER / O1_WAVES_PER_CYCLE)
#define O2_LOAD_STEP_RESOLUTION_W ((float)O2_CONNECTED_POWER / O2_WAVES_PER_CYCLE)
#define O3_LOAD_STEP_RESOLUTION_W ((float)O3_CONNECTED_POWER / O3_WAVES_PER_CYCLE)
#if (O1_CONNECTED_POWER == 0)
#define O1_LOAD_STEP_RESOLUTION_P 0.0
#else
#define O1_LOAD_STEP_RESOLUTION_P ((float)(O1_LOAD_STEP_RESOLUTION_W * 100.0) / O1_CONNECTED_POWER)
#endif
#if (O2_CONNECTED_POWER == 0)
#define O2_LOAD_STEP_RESOLUTION_P 0.0
#else
#define O2_LOAD_STEP_RESOLUTION_P ((float)(O2_LOAD_STEP_RESOLUTION_W * 100.0) / O2_CONNECTED_POWER)
#endif
#if (O3_CONNECTED_POWER == 0)
#define O3_LOAD_STEP_RESOLUTION_P 0.0
#else
#define O3_LOAD_STEP_RESOLUTION_P ((float)(O3_LOAD_STEP_RESOLUTION_W * 100.0) / O3_CONNECTED_POWER)
#endif
#define O1_LOAD_STEP_ENERGY ((float)O1_CONNECTED_POWER / (O1_WAVES_PER_CYCLE * 3600000 / O1_LOAD_STEP_CYCLE_TIME))
#define O2_LOAD_STEP_ENERGY ((float)O2_CONNECTED_POWER / (O2_WAVES_PER_CYCLE * 3600000 / O2_LOAD_STEP_CYCLE_TIME))
#define O3_LOAD_STEP_ENERGY ((float)O3_CONNECTED_POWER / (O3_WAVES_PER_CYCLE * 3600000 / O3_LOAD_STEP_CYCLE_TIME))
#define O1_PATTERN_VARS ((O1_WAVES_PER_CYCLE >> 6) + 1)
#define O2_PATTERN_VARS ((O2_WAVES_PER_CYCLE >> 6) + 1)
#define O3_PATTERN_VARS ((O3_WAVES_PER_CYCLE >> 6) + 1)

// Constants
const char *SOFTWARE_VERSION = "0.1.1";

// Variables
static volatile bool mains_frequency_ok = false;   // true = mains frequency is okay
static volatile bool o1_new_load_request = false;  // true = last (half) wave(s) in this load step cycle has/have begun... last chance to set a new load value
static volatile bool o2_new_load_request = false;
static volatile bool o3_new_load_request = false;
static volatile bool o1_new_load_cycle = false;  // true = a new load step cycle has started
static volatile bool o2_new_load_cycle = false;
static volatile bool o3_new_load_cycle = false;
#if (O1_OPERATING_MODE == 1)
static volatile bool o1_full_wave_indicator = false;  // true = the output state was taken for a full sine wave
#endif
#if (O2_OPERATING_MODE == 1)
static volatile bool o2_full_wave_indicator = false;
#endif
#if (O3_OPERATING_MODE == 1)
static volatile bool o3_full_wave_indicator = false;
#endif
static volatile bool k1_state = false;
static volatile bool k2_state = false;
static volatile uint32_t o1_load_value = 0;        // load value for output 1 (valid over an entire load cycle)
static volatile uint32_t o2_load_value = 0;        // load value for output 2 (valid over an entire load cycle)
static volatile uint32_t o3_load_value = 0;        // load value for output 3 (valid over an entire load cycle)
static volatile uint32_t o1_cycle_load_value = 0;  // load value cache for outout O1
static volatile uint32_t o2_cycle_load_value = 0;  // load value cache for outout O2
static volatile uint32_t o3_cycle_load_value = 0;  // load value cache for outout O3
static volatile uint32_t o1_wave_counter = 0;
static volatile uint32_t o2_wave_counter = 0;
static volatile uint32_t o3_wave_counter = 0;
static volatile uint32_t mains_stabilization_counter = 0;
static volatile uint32_t phase_trigger_counter = 0;  // 0 = phase 3 / 1 = phase 2 / 2 = phase 1
static volatile uint64_t half_wave_time = 0;
static volatile uint64_t o1_load_counter = { 0ULL };
static volatile uint64_t o2_load_counter = { 0ULL };
static volatile uint64_t o3_load_counter = { 0ULL };
static volatile uint64_t k1_load_counter = { 0ULL };
static volatile uint64_t k2_load_counter = { 0ULL };


#if (REMAINING_WAVES < 1)
#error Remaining (half) waves to request a new load value must be greater than 0
#endif
#if (O1_WAVES_PER_CYCLE <= REMAINING_WAVES)
#error Too many remaining (half) waves to request a new load value at output O1
#endif
#if (O2_WAVES_PER_CYCLE <= REMAINING_WAVES)
#error Too many remaining (half) waves to request a new load value at output O2
#endif
#if (O3_WAVES_PER_CYCLE <= REMAINING_WAVES)
#error Too many remaining (half) waves to request a new load value at output O3
#endif
#if (O1_LOAD_PATTERN == 1)
#if (((O1_WAVES_PER_CYCLE + 1) * O1_PATTERN_VARS) > MAXIMUM_CYCLE_PATTERN_VARS)
#error Insufficient memory space -> Please reduce load step resolution, or change load pattern to asymmetrical for output O1.
#endif
static uint64_t o1_load_pattern[O1_WAVES_PER_CYCLE + 1][O1_PATTERN_VARS] = { 0ULL };
#endif
#if (O2_LOAD_PATTERN == 1)
#if (((O2_WAVES_PER_CYCLE + 1) * O2_PATTERN_VARS) > MAXIMUM_CYCLE_PATTERN_VARS)
#error Insufficient memory space -> Please reduce load step resolution, or change load pattern to asymmetrical for output O2.
#endif
static uint64_t o2_load_pattern[O2_WAVES_PER_CYCLE + 1][O2_PATTERN_VARS] = { 0ULL };
#endif
#if (O3_LOAD_PATTERN == 1)
#if (((O3_WAVES_PER_CYCLE + 1) * O3_PATTERN_VARS) > MAXIMUM_CYCLE_PATTERN_VARS)
#error Insufficient memory space -> Please reduce load step resolution, or change load pattern to asymmetrical for output O3.
#endif
static uint64_t o3_load_pattern[O3_WAVES_PER_CYCLE + 1][O3_PATTERN_VARS] = { 0ULL };
#endif

bool t1 = true;
bool t2 = true;
bool t3 = true;
uint32_t v1 = 0;
uint32_t v2 = 0;
uint32_t v3 = 0;

hw_timer_t *hw_timer[4] = { NULL };

int32_t systick(void)  // Transfer mills() into signed integer, to prevent compare error at overrun
{
  return millis();
}


void error_handler(int error_number, int error_action) {
  Serial.println();
  Serial.print("Error ");
  Serial.print(error_number);
  Serial.print(": ");
  switch (error_number) {
    case 200:
      Serial.println("APB clock frequency is to low.");
      break;
    case 201:
      Serial.println("Could not configure hardware timer.");
      break;
    case 300:
      Serial.println("I2C -> A different number of bytes is available than requested");
      break;
    default:
      Serial.println("Unknown error.");
      error_action &= ~1;
      break;
  }
  if (error_action & 1) {
    Serial.println("ESP32 is restarting in 5 seconds...");
    delay(5000);
    ESP.restart();
  } else {
    while (1) {
      delay(100);
    }
  }
}



void stop_hw_timer(int timer_no) {
  if (hw_timer[timer_no]) {
    if (timerStarted(hw_timer[timer_no])) {
      timerStop(hw_timer[timer_no]);
    }
    if (timerAlarmEnabled(hw_timer[timer_no])) {
      timerAlarmDisable(hw_timer[timer_no]);
    }
  }
}


void init_hw_timer(int timer_no, uint64_t timer_value, bool cnt_up, bool timer_reload, uint64_t alarm_value, bool alarm_enable) {
  if (hw_timer[timer_no]) {
    stop_hw_timer(timer_no);
    timerDetachInterrupt(hw_timer[timer_no]);
    timerEnd(hw_timer[timer_no]);
    hw_timer[timer_no] = NULL;
  }
  hw_timer[timer_no] = timerBegin((uint16_t)(getApbFrequency() / TIME_BASE));
  if (hw_timer[timer_no] == NULL) error_handler(201, 1);
  timerStop(hw_timer[timer_no]);
  timerWrite(hw_timer[timer_no], timer_value);
  timerSetCountUp(hw_timer[timer_no], cnt_up);
  timerAlarm(hw_timer[timer_no], alarm_value, timer_reload, 0);
  if (alarm_enable) {
    timerAlarmEnable(hw_timer[timer_no]);
  } else {
    timerAlarmDisable(hw_timer[timer_no]);
  }
}


#if (O1_LOAD_PATTERN == 1)
bool o1_get_pattern(uint32_t load, uint32_t wave) {
  return (o1_load_pattern[load][wave >> 6] & (uint64_t)1 << (wave & 63ULL));
}

void o1_set_pattern(uint32_t load, uint32_t wave) {
  o1_load_pattern[load][wave >> 6] |= (uint64_t)1 << (wave & 63ULL);
}

void o1_load_pattern_init(void) {
  float quotient, half_quotient;
  uint32_t i, j, pattern;
  for (i = 1; i <= O1_WAVES_PER_CYCLE; i++) {
    quotient = (float)O1_WAVES_PER_CYCLE / i;
    half_quotient = quotient / 2.0;
    j = 0;
    do {
      pattern = (uint32_t)(half_quotient + (quotient * j));
      o1_set_pattern(i, pattern);
      j++;
    } while (j < i);
  }
}
#endif


#if (O2_LOAD_PATTERN == 1)
bool o2_get_pattern(uint32_t load, uint32_t wave) {
  return (o2_load_pattern[load][wave >> 6] & (uint64_t)1 << (wave & 63ULL));
}

void o2_set_pattern(uint32_t load, uint32_t wave) {
  o2_load_pattern[load][wave >> 6] |= (uint64_t)1 << (wave & 63ULL);
}

void o2_load_pattern_init(void) {
  float quotient, half_quotient;
  uint32_t i, j, pattern;
  for (i = 1; i <= O2_WAVES_PER_CYCLE; i++) {
    quotient = (float)O2_WAVES_PER_CYCLE / i;
    half_quotient = quotient / 2.0;
    j = 0;
    do {
      pattern = (uint32_t)(half_quotient + (quotient * j));
      o2_set_pattern(i, pattern);
      j++;
    } while (j < i);
  }
}
#endif


#if (O3_LOAD_PATTERN == 1)
bool o3_get_pattern(uint32_t load, uint32_t wave) {
  return (o3_load_pattern[load][wave >> 6] & (uint64_t)1 << (wave & 63ULL));
}

void o3_set_pattern(uint32_t load, uint32_t wave) {
  o3_load_pattern[load][wave >> 6] |= (uint64_t)1 << (wave & 63ULL);
}

void o3_load_pattern_init(void) {
  float quotient, half_quotient;
  uint32_t i, j, pattern;
  for (i = 1; i <= O3_WAVES_PER_CYCLE; i++) {
    quotient = (float)O3_WAVES_PER_CYCLE / i;
    half_quotient = quotient / 2.0;
    j = 0;
    do {
      pattern = (uint32_t)(half_quotient + (quotient * j));
      o3_set_pattern(i, pattern);
      j++;
    } while (j < i);
  }
}
#endif


// Zero crossing interrupt service routine
void IRAM_ATTR zero_corssing_detection_isr(void) {
  static volatile bool pre_trigger = false;
  portENTER_CRITICAL_ISR(&timerMux);
  half_wave_time = timerRead(hw_timer[0]);
  if (half_wave_time >= HALF_WAVE_DURATION_MIN) {
    timerWrite(hw_timer[0], 0ULL);  // reset sine half wave time out (hardware timer 0)
    if (half_wave_time <= HALF_WAVE_DURATION_MAX) {
      if (mains_frequency_ok) {
        timerWrite(hw_timer[1], 0ULL);                               // reset burst timer (hardware timer 1)
        timerAlarm(hw_timer[1], TRIGGER_ANGLE_PHASE_3, false, 0);  // assign burst timer to trigger angle phase 3
        timerStart(hw_timer[1]);
        timerAlarmEnable(hw_timer[1]);
        phase_trigger_counter = 0;


#if (O1_OPERATING_MODE == 1)
        if (o1_full_wave_indicator) {
#endif
          if (++o1_wave_counter == O1_WAVES_PER_CYCLE - REMAINING_WAVES) {
            o1_new_load_request = true;
          } else if (o1_wave_counter >= O1_WAVES_PER_CYCLE) {
            o1_load_counter += o1_cycle_load_value;
            o1_cycle_load_value = o1_load_value;
            if (o1_cycle_load_value > O1_WAVES_PER_CYCLE) {
              o1_cycle_load_value = O1_WAVES_PER_CYCLE;
            }
            o1_new_load_cycle = true;
            o1_wave_counter = 0;
          }
#if (O1_OPERATING_MODE == 1)
        }
        o1_full_wave_indicator = !o1_full_wave_indicator;
#endif


#if (O2_OPERATING_MODE == 1)
        if (o2_full_wave_indicator) {
#endif
          if (++o2_wave_counter == O2_WAVES_PER_CYCLE - REMAINING_WAVES) {
            o2_new_load_request = true;
          } else if (o2_wave_counter >= O2_WAVES_PER_CYCLE) {
            o2_load_counter += o2_cycle_load_value;
            o2_cycle_load_value = o2_load_value;
            if (o2_cycle_load_value > O2_WAVES_PER_CYCLE) {
              o2_cycle_load_value = O2_WAVES_PER_CYCLE;
            }
            o2_new_load_cycle = true;
            o2_wave_counter = 0;
          }
#if (O2_OPERATING_MODE == 1)
        }
        o2_full_wave_indicator = !o2_full_wave_indicator;
#endif


#if (O3_OPERATING_MODE == 1)
        if (o3_full_wave_indicator) {
#endif
          if (++o3_wave_counter == O3_WAVES_PER_CYCLE - REMAINING_WAVES) {
            o3_new_load_request = true;
          } else if (o3_wave_counter >= O3_WAVES_PER_CYCLE) {
            o3_load_counter += o3_cycle_load_value;
            o3_cycle_load_value = o3_load_value;
            if (o3_cycle_load_value > O3_WAVES_PER_CYCLE) {
              o3_cycle_load_value = O3_WAVES_PER_CYCLE;
            }
            o3_new_load_cycle = true;
            o3_wave_counter = 0;
          }
#if (O3_OPERATING_MODE == 1)
        }
        o3_full_wave_indicator = !o3_full_wave_indicator;
#endif

      } else {
        if (++mains_stabilization_counter >= MAINS_STABILIZATION_VALUE) {
          if (timerStarted(hw_timer[1])) {
            timerStop(hw_timer[1]);
          }
          timerWrite(hw_timer[1], 0ULL);                               // reset burst timer (hardware timer 1)
          timerAlarm(hw_timer[1], TRIGGER_ANGLE_PHASE_3, false, 0);  // assign burst timer to trigger angle phase 3
          timerStart(hw_timer[1]);
          timerAlarmEnable(hw_timer[1]);
          phase_trigger_counter = 0;
          o1_cycle_load_value = o1_load_value;
          if (o1_cycle_load_value > O1_WAVES_PER_CYCLE) {
            o1_cycle_load_value = O1_WAVES_PER_CYCLE;
          }
          o2_cycle_load_value = o2_load_value;
          if (o2_cycle_load_value > O2_WAVES_PER_CYCLE) {
            o2_cycle_load_value = O2_WAVES_PER_CYCLE;
          }
          o3_cycle_load_value = o3_load_value;
          if (o3_cycle_load_value > O3_WAVES_PER_CYCLE) {
            o3_cycle_load_value = O3_WAVES_PER_CYCLE;
          }
          o1_wave_counter = 0;
          o2_wave_counter = 0;
          o3_wave_counter = 0;
#if (O1_OPERATING_MODE == 1)
          o1_full_wave_indicator = false;
#endif
#if (O2_OPERATING_MODE == 1)
          o2_full_wave_indicator = false;
#endif
#if (O3_OPERATING_MODE == 1)
          o3_full_wave_indicator = false;
#endif
          mains_frequency_ok = true;
        }
      }
    } else {
      if (mains_frequency_ok) {
        if (timerStarted(hw_timer[1])) {
          timerStop(hw_timer[1]);
        }
        mains_frequency_ok = false;
        digitalWrite(O1_OUTPUT, LOW);
        digitalWrite(O2_OUTPUT, LOW);
        digitalWrite(O3_OUTPUT, LOW);
      }
      mains_stabilization_counter = 0;
    }
    pre_trigger = false;
  } else {
    if (pre_trigger) {
      timerWrite(hw_timer[0], 0ULL);  // reset sine half wave time out (hardware timer 0)
      if (mains_frequency_ok) {
        if (timerStarted(hw_timer[1])) {
          timerStop(hw_timer[1]);
        }
        mains_frequency_ok = false;
        digitalWrite(O1_OUTPUT, LOW);
        digitalWrite(O2_OUTPUT, LOW);
        digitalWrite(O3_OUTPUT, LOW);
      }
      mains_stabilization_counter = 0;
    } else {
      pre_trigger = true;
    }
  }
  portEXIT_CRITICAL_ISR(&timerMux);
}


// Burst trigger interrupt service routine (hardware timer 1)
void IRAM_ATTR burst_trigger_isr(void) {
  portENTER_CRITICAL_ISR(&timerMux);
  if (mains_frequency_ok) {
    switch (phase_trigger_counter) {
      case 0:                                                        // set burst trigger phase 3 and reset burst trigger phase 1
        timerAlarm(hw_timer[1], TRIGGER_ANGLE_PHASE_2, false, 0);  // assign burst timer to trigger angle phase 2
#if (O1_ASSIGNED_TO_PHASE == 1)
        digitalWrite(O1_OUTPUT, LOW);
#endif
#if (O1_ASSIGNED_TO_PHASE == 3)
#if (O1_LOAD_PATTERN == 1)
        if (o1_get_pattern(o1_cycle_load_value, o1_wave_counter)) {
          digitalWrite(O1_OUTPUT, HIGH);
        }
#else
        if (o1_wave_counter < o1_cycle_load_value) {
          digitalWrite(O1_OUTPUT, HIGH);
        }
#endif
#endif

#if (O2_ASSIGNED_TO_PHASE == 1)
        digitalWrite(O2_OUTPUT, LOW);
#endif
#if (O2_ASSIGNED_TO_PHASE == 3)
#if (O2_LOAD_PATTERN == 1)
        if (o2_get_pattern(o2_cycle_load_value, o2_wave_counter)) {
          digitalWrite(O2_OUTPUT, HIGH);
        }
#else
        if (o2_wave_counter < o2_cycle_load_value) {
          digitalWrite(O2_OUTPUT, HIGH);
        }
#endif
#endif

#if (O3_ASSIGNED_TO_PHASE == 1)
        digitalWrite(O3_OUTPUT, LOW);
#endif
#if (O3_ASSIGNED_TO_PHASE == 3)
#if (O3_LOAD_PATTERN == 1)
        if (o3_get_pattern(o3_cycle_load_value, o3_wave_counter)) {
          digitalWrite(O3_OUTPUT, HIGH);
        }
#else
        if (o3_wave_counter < o3_cycle_load_value) {
          digitalWrite(O3_OUTPUT, HIGH);
        }
#endif
#endif
        break;

      case 1:                                                        // set burst trigger phase 2 and reset burst trigger phase 3
        timerAlarm(hw_timer[1], TRIGGER_ANGLE_PHASE_1, false, 0);  // assign burst timer to trigger angle phase 1
#if (O1_ASSIGNED_TO_PHASE == 3)
        digitalWrite(O1_OUTPUT, LOW);
#endif
#if (O1_ASSIGNED_TO_PHASE == 2)
#if (O1_LOAD_PATTERN == 1)
        if (o1_get_pattern(o1_cycle_load_value, o1_wave_counter)) {
          digitalWrite(O1_OUTPUT, HIGH);
        }
#else
        if (o1_wave_counter < o1_cycle_load_value) {
          digitalWrite(O1_OUTPUT, HIGH);
        }
#endif
#endif

#if (O2_ASSIGNED_TO_PHASE == 3)
        digitalWrite(O2_OUTPUT, LOW);
#endif
#if (O2_ASSIGNED_TO_PHASE == 2)
#if (O2_LOAD_PATTERN == 1)
        if (o2_get_pattern(o2_cycle_load_value, o2_wave_counter)) {
          digitalWrite(O2_OUTPUT, HIGH);
        }
#else
        if (o2_wave_counter < o2_cycle_load_value) {
          digitalWrite(O2_OUTPUT, HIGH);
        }
#endif
#endif

#if (O3_ASSIGNED_TO_PHASE == 3)
        digitalWrite(O3_OUTPUT, LOW);
#endif
#if (O3_ASSIGNED_TO_PHASE == 2)
#if (O3_LOAD_PATTERN == 1)
        if (o3_get_pattern(o3_cycle_load_value, o3_wave_counter)) {
          digitalWrite(O3_OUTPUT, HIGH);
        }
#else
        if (o3_wave_counter < o3_cycle_load_value) {
          digitalWrite(O3_OUTPUT, HIGH);
        }
#endif
#endif
        break;

      case 2:  // set burst trigger phase 1 and reset burst trigger phase 2
        timerStop(hw_timer[1]);
#if (O1_ASSIGNED_TO_PHASE == 2)
        digitalWrite(O1_OUTPUT, LOW);
#endif
#if (O1_ASSIGNED_TO_PHASE == 1)
#if (O1_LOAD_PATTERN == 1)
        if (o1_get_pattern(o1_cycle_load_value, o1_wave_counter)) {
          digitalWrite(O1_OUTPUT, HIGH);
        }
#else
        if (o1_wave_counter < o1_cycle_load_value) {
          digitalWrite(O1_OUTPUT, HIGH);
        }
#endif
#endif

#if (O2_ASSIGNED_TO_PHASE == 2)
        digitalWrite(O2_OUTPUT, LOW);
#endif
#if (O2_ASSIGNED_TO_PHASE == 1)
#if (O2_LOAD_PATTERN == 1)
        if (o2_get_pattern(o2_cycle_load_value, o2_wave_counter)) {
          digitalWrite(O2_OUTPUT, HIGH);
        }
#else
        if (o2_wave_counter < o2_cycle_load_value) {
          digitalWrite(O2_OUTPUT, HIGH);
        }
#endif
#endif

#if (O3_ASSIGNED_TO_PHASE == 2)
        digitalWrite(O3_OUTPUT, LOW);
#endif
#if (O3_ASSIGNED_TO_PHASE == 1)
#if (O3_LOAD_PATTERN == 1)
        if (o3_get_pattern(o3_cycle_load_value, o3_wave_counter)) {
          digitalWrite(O3_OUTPUT, HIGH);
        }
#else
        if (o3_wave_counter < o3_cycle_load_value) {
          digitalWrite(O3_OUTPUT, HIGH);
        }
#endif
#endif
        break;
    }
    ++phase_trigger_counter;
  }
  portEXIT_CRITICAL_ISR(&timerMux);
}


// sine half wave time out interrupt service routine (hardware timer 0)
void IRAM_ATTR half_wave_time_out_isr(void) {
  portENTER_CRITICAL_ISR(&timerMux);
  if (mains_frequency_ok) {
    if (timerStarted(hw_timer[1])) {
      timerStop(hw_timer[1]);
    }
    mains_frequency_ok = false;
    digitalWrite(O1_OUTPUT, LOW);
    digitalWrite(O2_OUTPUT, LOW);
    digitalWrite(O3_OUTPUT, LOW);
  }
  mains_stabilization_counter = 0;
  portEXIT_CRITICAL_ISR(&timerMux);
}



bool mains_frequnecy_okay(void) {
  bool mfok;
  portENTER_CRITICAL(&timerMux);
  mfok = mains_frequency_ok;
  portEXIT_CRITICAL(&timerMux);
  return mfok;
}



uint64_t get_half_wave_time(void) {
  uint64_t hwt;
  if (mains_frequnecy_okay()) {
    portENTER_CRITICAL(&timerMux);
    hwt = half_wave_time;
    portEXIT_CRITICAL(&timerMux);
  } else {
    hwt = 0ULL;
  }
  return hwt;
}

float get_mains_frequency(void) {
  uint64_t hwt = get_half_wave_time();
  if (hwt > 0ULL) {
    return (float)(TIME_BASE >> 1) / hwt;
  }
  return 0.0;
}


bool can_i_set_new_load_value(uint32_t output) {
  bool nlrq;
  switch (output) {
    case 1:
      portENTER_CRITICAL(&timerMux);
      nlrq = o1_new_load_request;
      if (nlrq) {
        o1_new_load_request = false;
      }
      portEXIT_CRITICAL(&timerMux);
      break;
    case 2:
      portENTER_CRITICAL(&timerMux);
      nlrq = o2_new_load_request;
      if (nlrq) {
        o2_new_load_request = false;
      }
      portEXIT_CRITICAL(&timerMux);
      break;
    case 3:
      portENTER_CRITICAL(&timerMux);
      nlrq = o3_new_load_request;
      if (nlrq) {
        o3_new_load_request = false;
      }
      portEXIT_CRITICAL(&timerMux);
      break;
    default:
      nlrq = false;
      break;
  }
  return nlrq;
}


bool new_load_cycle_started(uint32_t output) {
  bool nlc;
  switch (output) {
    case 1:
      portENTER_CRITICAL(&timerMux);
      nlc = o1_new_load_cycle;
      if (nlc) {
        o1_new_load_cycle = false;
      }
      portEXIT_CRITICAL(&timerMux);
      break;
    case 2:
      portENTER_CRITICAL(&timerMux);
      nlc = o2_new_load_cycle;
      if (nlc) {
        o2_new_load_cycle = false;
      }
      portEXIT_CRITICAL(&timerMux);
      break;
    case 3:
      portENTER_CRITICAL(&timerMux);
      nlc = o3_new_load_cycle;
      if (nlc) {
        o3_new_load_cycle = false;
      }
      portEXIT_CRITICAL(&timerMux);
      break;
    default:
      nlc = false;
      break;
  }
  return nlc;
}


void set_output_k1(void) {
  digitalWrite(K1_OUTPUT, HIGH);
  k1_state = true;
}

void reset_output_k1(void) {
  digitalWrite(K1_OUTPUT, LOW);
  k1_state = false;
}

void set_output_k2(void) {
  digitalWrite(K2_OUTPUT, HIGH);
  k2_state = true;
}

void reset_output_k2(void) {
  digitalWrite(K2_OUTPUT, LOW);
  k2_state = false;
}


float get_energy_wh(uint32_t output)  // Get energy of an output in watt hours
{
  uint64_t lc;
  float energy = 0.0;
  switch (output) {
    case 1:
      portENTER_CRITICAL(&timerMux);
      lc = o1_load_counter;
      portEXIT_CRITICAL(&timerMux);
      energy = (float)lc * O1_LOAD_STEP_ENERGY;
      break;
    case 2:
      portENTER_CRITICAL(&timerMux);
      lc = o2_load_counter;
      portEXIT_CRITICAL(&timerMux);
      energy = (float)lc * O2_LOAD_STEP_ENERGY;
      break;
    case 3:
      portENTER_CRITICAL(&timerMux);
      lc = o3_load_counter;
      portEXIT_CRITICAL(&timerMux);
      energy = (float)lc * O3_LOAD_STEP_ENERGY;
      break;
    default:
      energy = 0.0;
      break;
  }
  return energy;
}

float get_energy_kwh(uint32_t output)  // Get energy of an output in kilowatt hours
{
  uint64_t lc;
  float energy = 0.0;
  switch (output) {
    case 1:
      portENTER_CRITICAL(&timerMux);
      lc = o1_load_counter;
      portEXIT_CRITICAL(&timerMux);
      energy = ((float)lc / 1000.0) * O1_LOAD_STEP_ENERGY;
      break;
    case 2:
      portENTER_CRITICAL(&timerMux);
      lc = o2_load_counter;
      portEXIT_CRITICAL(&timerMux);
      energy = ((float)lc / 1000.0) * O3_LOAD_STEP_ENERGY;
      break;
    case 3:
      portENTER_CRITICAL(&timerMux);
      lc = o3_load_counter;
      portEXIT_CRITICAL(&timerMux);
      energy = ((float)lc / 1000.0) * O3_LOAD_STEP_ENERGY;
      break;
    default:
      energy = 0.0;
      break;
  }
  return energy;
}


void outputs_init(void) {
  digitalWrite(O1_OUTPUT, LOW);
  digitalWrite(O2_OUTPUT, LOW);
  digitalWrite(O3_OUTPUT, LOW);
  reset_output_k1();
  reset_output_k2();
  pinMode(O1_OUTPUT, OUTPUT);
  pinMode(O2_OUTPUT, OUTPUT);
  pinMode(O3_OUTPUT, OUTPUT);
  pinMode(K1_OUTPUT, OUTPUT);
  pinMode(K2_OUTPUT, OUTPUT);
  pinMode(ZERO_CROSSING_DETECTION_PHW, INPUT);
  pinMode(ZERO_CROSSING_DETECTION_NHW, INPUT);
}


void burst_control_init(void) {
  Serial.println("Load controller configuration...");
  Serial.print("Software version ");
  Serial.println(SOFTWARE_VERSION);
  Serial.print("Mains rated frequency: ");
  Serial.print(MAINS_FREQUENCY);
  Serial.println(" Hz");
  Serial.print("Allowed frequency deviation: ");
  Serial.print(MIN_FREQUENCY_DEVIATION, 1);
  Serial.print(" - ");
  Serial.print(MAX_FREQUENCY_DEVIATION, 1);
  Serial.print(" Hz (±");
  Serial.print(MAXIMUM_FREQUENCY_DEVIATION);
  Serial.println("%)");

  Serial.println();
  Serial.println("Digital output O1:");
  Serial.print("- Connected Power: ");
  Serial.print(O1_CONNECTED_POWER);
  Serial.print(" W at phase ");
  Serial.print(O1_ASSIGNED_TO_PHASE);
  if (O1_LOAD_PATTERN == 0) {
    Serial.println(" with asymmetrical load pattern");
  } else {
    Serial.println(" with symmetrical load pattern");
  }
  Serial.print("- Load step resolution: ");
  Serial.print(O1_LOAD_STEP_RESOLUTION_W, 2);
  Serial.print(" W (");
  Serial.print(O1_LOAD_STEP_RESOLUTION_P, 1);
  Serial.println("%)");
#if (O1_OPERATING_MODE == 0)
  Serial.print("- Half sine waves per load step cycle (Full load value): ");
#else
  Serial.print("- Full sine waves per load step cycle (Full load value): ");
#endif
  Serial.println(O1_WAVES_PER_CYCLE);
  Serial.print("- Load step cycle time: ");
  Serial.print(O1_LOAD_STEP_CYCLE_TIME);
  Serial.print(" ms (");
  Serial.print(O1_LOAD_STEP_RESOLUTION);
  Serial.println("%)");

  Serial.println();
  Serial.println("Digital output O2:");
  Serial.print("- Connected Power: ");
  Serial.print(O2_CONNECTED_POWER);
  Serial.print(" W at phase ");
  Serial.print(O2_ASSIGNED_TO_PHASE);
  if (O2_LOAD_PATTERN == 0) {
    Serial.println(" with asymmetrical load pattern");
  } else {
    Serial.println(" with symmetrical load pattern");
  }
  Serial.print("- Load step resolution: ");
  Serial.print(O2_LOAD_STEP_RESOLUTION_W, 2);
  Serial.print(" W (");
  Serial.print(O2_LOAD_STEP_RESOLUTION_P, 1);
  Serial.println("%)");
#if (O2_OPERATING_MODE == 0)
  Serial.print("- Half sine waves per load step cycle (Full load value): ");
#else
  Serial.print("- Full sine waves per load step cycle (Full load value): ");
#endif
  Serial.println(O2_WAVES_PER_CYCLE);
  Serial.print("- Load step cycle time: ");
  Serial.print(O2_LOAD_STEP_CYCLE_TIME);
  Serial.print(" ms (");
  Serial.print(O2_LOAD_STEP_RESOLUTION);
  Serial.println("%)");

  Serial.println();
  Serial.println("Digital output O3:");
  Serial.print("- Connected Power: ");
  Serial.print(O3_CONNECTED_POWER);
  Serial.print(" W at phase ");
  Serial.print(O3_ASSIGNED_TO_PHASE);
  if (O3_LOAD_PATTERN == 0) {
    Serial.println(" with asymmetrical load pattern");
  } else {
    Serial.println(" with symmetrical load pattern");
  }
  Serial.print("- Load step resolution: ");
  Serial.print(O3_LOAD_STEP_RESOLUTION_W, 2);
  Serial.print(" W (");
  Serial.print(O3_LOAD_STEP_RESOLUTION_P, 1);
  Serial.println("%)");
#if (O3_OPERATING_MODE == 0)
  Serial.print("- Half sine waves per load step cycle (Full load value): ");
#else
  Serial.print("- Full sine waves per load step cycle (Full load value): ");
#endif
  Serial.println(O3_WAVES_PER_CYCLE);
  Serial.print("- Load step cycle time: ");
  Serial.print(O3_LOAD_STEP_CYCLE_TIME);
  Serial.print(" ms (");
  Serial.print(O3_LOAD_STEP_RESOLUTION);
  Serial.println("%)");

  Serial.println();
  Serial.println("Relay output K1:");
  Serial.print("- Connected Power: ");
  Serial.print(K1_CONNECTED_POWER);
  Serial.println(" W");

  Serial.println();
  Serial.println("Relay output K2:");
  Serial.print("- Connected Power: ");
  Serial.print(K2_CONNECTED_POWER);
  Serial.println(" W");

  if (getApbFrequency() < (TIME_BASE + TIME_BASE)) error_handler(200, 0);

#if (O1_LOAD_PATTERN == 1)
  o1_load_pattern_init();
#endif
#if (O2_LOAD_PATTERN == 1)
  o2_load_pattern_init();
#endif
#if (O3_LOAD_PATTERN == 1)
  o3_load_pattern_init();
#endif

  init_hw_timer(1, 0ULL, true, false, TRIGGER_ANGLE_PHASE_3, true);
  timerAttachInterrupt(hw_timer[1], &burst_trigger_isr);

  init_hw_timer(0, 0ULL, true, true, TRIGGER_ANGLE_TIME_OUT, true);
  timerAttachInterrupt(hw_timer[0], &half_wave_time_out_isr);
  timerStart(hw_timer[0]);

  attachInterrupt(digitalPinToInterrupt(ZERO_CROSSING_DETECTION_PHW), zero_corssing_detection_isr, RISING);
  attachInterrupt(digitalPinToInterrupt(ZERO_CROSSING_DETECTION_NHW), zero_corssing_detection_isr, RISING);
  Serial.println("Burst control service has started...");
}



void setup(void) {
  uint32_t i, j;

  outputs_init();


  pinMode(36, INPUT);
  pinMode(39, INPUT);

  Serial.begin(115200);
  while (!Serial)
    ;
  delay(10);
  Serial.println();

  Serial.print("O1_LOAD_STEP_ENERGY: ");
  Serial.println(O1_LOAD_STEP_ENERGY, 8);
  Serial.print("O2_LOAD_STEP_ENERGY: ");
  Serial.println(O2_LOAD_STEP_ENERGY, 8);
  Serial.print("O3_LOAD_STEP_ENERGY: ");
  Serial.println(O3_LOAD_STEP_ENERGY, 8);
  burst_control_init();


#if (O1_LOAD_PATTERN == 1)
  Serial.println("O1 - Pattern");
  for (i = 0; i <= O1_WAVES_PER_CYCLE; i++) {
    Serial.print(i);
    Serial.print("- ");
    for (j = 0; j < O1_WAVES_PER_CYCLE; j++) {
      if (o1_get_pattern(i, j)) {
        Serial.print("1");
      } else
        Serial.print("0");
    }
    Serial.println();
  }
#endif

#if (O2_LOAD_PATTERN == 1)
  Serial.println("O2 - Pattern");
  for (i = 0; i <= O2_WAVES_PER_CYCLE; i++) {
    Serial.print(i);
    Serial.print("- ");
    for (j = 0; j < O2_WAVES_PER_CYCLE; j++) {
      if (o2_get_pattern(i, j)) {
        Serial.print("1");
      } else
        Serial.print("0");
    }
    Serial.println();
  }
#endif

#if (O3_LOAD_PATTERN == 1)
  Serial.println("O3 - Pattern");
  for (i = 0; i <= O3_WAVES_PER_CYCLE; i++) {
    Serial.print(i);
    Serial.print("- ");
    for (j = 0; j < O3_WAVES_PER_CYCLE; j++) {
      if (o3_get_pattern(i, j)) {
        Serial.print("1");
      } else
        Serial.print("0");
    }
    Serial.println();
  }
#endif
}


void loop(void) {
  uint32_t z = systick();
  o1_load_value = 25;
  o2_load_value = 50;
  o3_load_value = 75;
  delay(11000);
  z = systick();
  while (1) {
    if (systick() - z > 999) {
      z = systick();
      for (int i = 1; i < 4; i++) {
        Serial.print("O");
        Serial.print(i);
        Serial.print(": ");
        Serial.print(get_energy_wh(i), 4);
        Serial.print(" Wh ");
        Serial.print(" - ");
      }
      Serial.println();
    }
  }
}
